import numpy as np
from findiff.coefs import coefficients, coefficients_non_uni


class FinDiff(object):
    """Wrapper class for finite difference linear differential operators in any number of dimensions
       on uniform and non-uniform grids."""

    def __init__(self, dims=[0], **kwargs):
        """Constructor for FinDiff class
        
            Parameters:
            -----------
            
            dims        list / array-like
                        A list with the dimensions along which to take derivatives. Repetition means
                        higher derivative. For instance, dims=[0] means first derivative along 0-th
                        axis, dims=[0,0,1] means third partial derivative, twice along 0-th axis, once
                        along 1st axis
            
            **kwargs:
            
                h       list / array-like
                        A list of real numbers with the grid spacings along all the axes in case
                        of uniform grids. The length of the list signifies the dimension, i.e. number
                        of independent variables. 
                
                coords  list / array-like
                        A list of 1D-arrays of real numbers with the coordinate values along each axis.
                        This signifies that you are using a non-uniform grid.
                        
                acc     even integer
                        The desired accuracy order. Default is acc=2."""

        if "h" in kwargs:   # we have a uniform grid
            if "coords" in kwargs:
                raise Exception("Either specify h or coords, not both.")
            h = kwargs["h"]
            self.uniform = True
        elif "coords" in kwargs:  # we have a non-uniform grid
            if "h" in kwargs:
                raise Exception("Either specify h or coords, not both.")
            coords = kwargs["coords"]
            self.uniform = False
        else:
            if "empty" not in kwargs:
                raise Exception("Neither h nor coords specified.")


        if "acc" in kwargs:
            acc = kwargs["acc"]
        else:
            acc = 2

        if "empty" in kwargs and kwargs["empty"]:
            self._basic_ops = []
            self._coefs = []
        else:
            if self.uniform:
                self._basic_ops = [BasicFinDiff(h, dims, acc)]
            else:
                self._basic_ops = [BasicFinDiffNonUniform(coords, dims, acc)]

            self._coefs = [Coefficient(1)]

    def __call__(self, y):
        """Applies the linear differential operator to y
        
            Parameters:
            -----------
            
                y       ndarray
                        The array to differentiate
                        
            Returns:
            --------
            
                An ndarray with the derivative. It has the same shape as y. """

        result = np.zeros_like(y)

        for c, op in zip(self._coefs, self._basic_ops):

            if isinstance(c.value, np.ndarray) or c.value != 1:
                result += c.value * op(y)
            else:
                result += op(y)

        return result

    def __add__(self, other):

        if self._grids_are_incompatible(other):
            raise ValueError("Operators on incompatible grids cannot be added.")

        new_op = FinDiff(empty=True)
        new_op.uniform = self.uniform
        new_op._basic_ops.extend(self._basic_ops)
        new_op._coefs.extend(self._coefs)
        new_op._basic_ops.extend(other._basic_ops)
        new_op._coefs.extend(other._coefs)
        return new_op

    def __mul__(self, other):

        if not isinstance(other, Coefficient):
            other = Coefficient(other)

        return other * self

    def __rmul__(self, other):

        if not isinstance(other, Coefficient):
            other = Coefficient(other)

        new_op = FinDiff(empty=True)
        new_op.uniform = self.uniform
        new_op._basic_ops.extend(self._basic_ops)
        new_op._coefs.extend(self._coefs)

        for i in range(len(new_op._coefs)):
            new_op._coefs[i].value *= other.value

        return new_op

    def _grids_are_incompatible(self, other):
        if self.uniform and other.uniform:
            return (self._basic_ops[0]._h != other._basic_ops[0]._h).any()
        if not self.uniform and not other.uniform:
            coords1 = self._basic_ops[0]._coords
            coords2 = other._basic_ops[0]._coords
            return (coords1 != coords2).any()
        return True


class BasicFinDiff(object):
    """Finite difference derivative of any order, any accuracy in any dimension for uniform grids 
    """

    def __init__(self, h=[1.], dims=[0], acc=2):
        """Constructor for Finite Difference operator on _uniform_ grids
           
           Parameters:
           ----------
           
           h        array-like
                    The grid spacing along each axis.
           dims     array-like
                    The axes along which to take the derivatives. Multiple values mean higher derivative.
           acc      int
                    The accuracy order of the finite difference scheme.
                            
           *Note*: You can use this class, but it is usually better to use the wrapper class FinDiff 
                            
           Example:
           --------
           
           Suppose f is a four-dimensional array. The second partial derivative with respect to the second axis,
                    
                    \frac{\partial^2 f}{\partial y^2},
                     
           on a grid with equidistant spacing h=[0.1, 0.1, 0.1, 0.1] is given by
           
                FinDiff(h=[0.1, 0.1, 0.1, 0.1], dims=[1, 1])
                
        """

        self._h = _wrap_in_array(h)
        self._dims = _wrap_in_array(dims)
        self._acc = acc

        ndims = len(self._h)
        self._derivs = [np.sum(self._dims == i) for i in range(ndims)]
        self._coefs = self._det_coefs()

    def _det_coefs(self):
        """Calculates the finite difference coefficients for the requested partial derivatives"""

        coefs = []

        for i in range(len(self._h)):
            coefs.append(coefficients(self._derivs[i], self._acc))

        return coefs

    def __call__(self, y):
        """Applies the finite difference operator to a function y"""

        ndims = len(y.shape)
        if ndims != len(self._h):
            raise IndexError("y and h have different dimensions")

        yd = np.array(y)

        for i in range(ndims):
            if self._derivs[i] > 0:
                yd = _diff_general(yd, self._h, self._derivs[i], i, self._acc, self._coefs[i])

        return yd


class BasicFinDiffNonUniform(object):
    """Finite difference derivative of any order, any accuracy in any dimension for uniform grids 
    """

    def __init__(self, coords, dims=[0], acc=2):

        self._coords = np.array(coords)
        self._dims = _wrap_in_array(dims)
        self._acc = acc

        ndims = len(self._coords)
        self._derivs = [np.sum(self._dims == i) for i in range(ndims)]
        self._coefs = self._det_coefs()

    def _det_coefs(self):
        """Calculates the finite difference coefficients for the requested partial derivatives"""

        coefs = []

        ndims = len(self._coords)

        for idim in range(ndims):
            c = []
            for i in range(len(self._coords[idim])):
                c.append(coefficients_non_uni(self._derivs[idim], self._acc, self._coords[idim], i))
            coefs.append(c)

        return coefs

    def __call__(self, y):

        ndims = len(y.shape)
        if ndims != len(self._coords):
            raise IndexError("y and h have different dimensions")

        yd = np.array(y)

        for idim in range(ndims):
            if self._derivs[idim] > 0:
                yd = _diff_general_non_uni(yd, self._coords, idim, self._coefs[idim])

        return yd


class Coefficient(object):
    """Encapsulates a constant (number) or variable (coordinate array) value to multiply with a linear operator
    """

    def __init__(self, value):
        self.value = value


class Laplacian(object):
    """A representation of the Laplace operator in arbitrary dimensions using finite difference schemes"""

    def __init__(self, h=[1.], acc=2):
        """Constructor for the Laplacian
        
           Parameters:
           -----------
           
           h        array-like
                    The grid spacing along each axis
           acc      int
                    The accuracy order of the finite difference scheme        
        """

        h = _wrap_in_array(h)

        self._parts = [FinDiff(h=h, dims=[k, k], acc=acc) for k in range(len(h))]

    def __call__(self, f):
        """Applies the Laplacian to the array f
        
           Parameters:
           -----------
           
           f        ndarray
                    The function to differentiate given as an array.
        
           Returns:
           --------    
           
           an ndarray with Laplace(f)
        
        """
        laplace_f = np.zeros_like(f)

        for part in self._parts:
            laplace_f += part(f)

        return laplace_f


def _diff_general_non_uni(y, coords, dim, coefs):
    """The core function to take a partial derivative on a non-uniform grid"""

    yd = np.zeros_like(y)

    ndims = len(y.shape)
    multi_slice = [slice(None, None)] * ndims
    ref_multi_slice = [slice(None, None)] * ndims

    for i, x in enumerate(coords[dim]):
        weights = coefs[i]["coefficients"]
        offsets = coefs[i]["offsets"]
        ref_multi_slice[dim] = i

        for off, w in zip(offsets, weights):
            multi_slice[dim] = i + off
            yd[ref_multi_slice] += w * y[multi_slice]

    return yd


def _diff_general(y, h, deriv, dim, acc, coefs=None):
    """The core function to take a partial derivative on a uniform grid.
    """

    if coefs is None:
        coefs = coefficients(deriv, acc)

    npts = y.shape[dim]

    scheme = "center"
    weights = coefs[scheme]["coefficients"]
    offsets = coefs[scheme]["offsets"]

    nbndry = len(weights) // 2
    ref_slice = slice(nbndry, npts - nbndry, 1)
    off_slices = [_shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))]

    yd = np.zeros_like(y)

    _apply_to_array(yd, y, weights, off_slices, ref_slice, dim)

    scheme = "forward"
    weights = coefs[scheme]["coefficients"]
    offsets = coefs[scheme]["offsets"]

    ref_slice = slice(0, nbndry, 1)
    off_slices = [_shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))]

    _apply_to_array(yd, y, weights, off_slices, ref_slice, dim)

    scheme = "backward"
    weights = coefs[scheme]["coefficients"]
    offsets = coefs[scheme]["offsets"]

    ref_slice = slice(npts - nbndry, npts, 1)
    off_slices = [_shift_slice(ref_slice, offsets[k], npts) for k in range(len(offsets))]

    _apply_to_array(yd, y, weights, off_slices, ref_slice, dim)

    h_inv = 1./h**deriv
    return yd * h_inv[dim]


def _apply_to_array(yd, y, weights, off_slices, ref_slice, dim):
    """Applies the finite differences only to slices along a given axis"""

    ndims = len(y.shape)

    all = slice(None, None, 1)

    ref_multi_slice = [all] * ndims
    ref_multi_slice[dim] = ref_slice

    for w, s in zip(weights, off_slices):
        off_multi_slice = [all] * ndims
        off_multi_slice[dim] = s
        yd[ref_multi_slice] += w * y[off_multi_slice]


def _shift_slice(sl, off, max_index):

    if sl.start + off < 0 or sl.stop + off > max_index:
        raise IndexError("Shift slice out of bounds")

    return slice(sl.start + off, sl.stop + off, sl.step)


def _wrap_in_array(val):

    if hasattr(val, "__len__"):
        return np.array(val)

    return np.array([val])



import findiff as fd
import unittest
import numpy as np


class TestUtilities(object):

    def _prep_1d_func(self, x0=-5, x1=5, nx=100):
        x = np.linspace(x0, x1, nx)
        dx = x[1] - x[0]
        f = x**3
        fxe = 3*x**2
        fxxe = 6*x
        return dx, f, fxe, fxxe

    def _assertAlmostEqual(self, f1, f2, tol=7):
        err = np.max(np.abs(f1-f2))
        self.assertAlmostEqual(0, err, tol)


class TestFinDiffUniform(unittest.TestCase, TestUtilities):

    def test_diff_1d_ord1_acc2(self):
        x = np.linspace(-3, 3, 50)
        h = x[1] - x[0]
        y = x ** 2
        y1e = 2 * x
        df = fd.FinDiff(h=h, dims=[0], acc=2)
        y1 = df(y)
        err = np.max(np.abs(y1e - y1))
        self.assertAlmostEqual(0, err)

    def test_diff_1d_ord1_acc4(self):
        x = np.linspace(-3, 3, 50)
        h = x[1] - x[0]
        y = x ** 2
        y1e = 2 * x
        y1 = fd.FinDiff(h=h, dims=[0], acc=4)(y)
        err = np.max(np.abs(y1e - y1))
        self.assertAlmostEqual(0, err)

    def test_diff_1d_ord2_acc2(self):
        x = np.linspace(-3, 3, 50)
        h = x[1] - x[0]
        y = x ** 3
        y2e = 6 * x
        y2 = fd.FinDiff(h=h, dims=[0, 0], acc=2)(y)
        err = np.max(np.abs(y2e - y2))
        self.assertAlmostEqual(0, err)

    def test_diff_1d_ord2_acc4(self):
        x = np.linspace(-3, 3, 50)
        h = x[1] - x[0]
        y = x ** 3
        y2e = 6 * x
        y2 = fd.FinDiff(h=h, dims=[0, 0], acc=4)(y)
        err = np.max(np.abs(y2e - y2))
        self.assertAlmostEqual(0, err)

    def test_diff_2d_ord1_acc2(self):
        x = np.linspace(-1, 1, 500)
        y = np.linspace(-1, 1, 500)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = Y**3 * X**3
        fxe = 3 * X**2 * Y**3
        fx = fd.FinDiff(h=[dx, dy], dims=[0], acc=2)(f)
        err = np.max(np.abs(fxe - fx))
        self.assertAlmostEqual(0, err, 4)
        fye = 3 * X**3 * Y**2
        fy = fd.FinDiff(h=[dx, dy], dims=[1], acc=2)(f)
        err = np.max(np.abs(fye - fy))
        self.assertAlmostEqual(0, err, 4)

    def test_diff_2d_ord2_acc2(self):
        x = np.linspace(-1, 1, 500)
        y = np.linspace(-1, 1, 500)
        dx = dy = x[1] - x[0]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = Y**3 * X**3
        fxxe = 6 * X * Y**3
        fxx = fd.FinDiff(h=[dx, dy], dims=[0, 0], acc=2)(f)
        err = np.max(np.abs(fxxe - fxx))
        self.assertAlmostEqual(0, err, 4)
        fyye = 6 * X**3 * Y
        fyy = fd.FinDiff(acc=2, h=[dx, dy], dims=[1, 1])(f)
        err = np.max(np.abs(fyye - fyy))
        self.assertAlmostEqual(0, err, 6)

    def test_diff_3d_ord1_acc2(self):
        x = np.linspace(-0.5, 0.5, 100)
        y = np.linspace(-0.5, 0.5, 100)
        z = np.linspace(-0.5, 0.5, 100)
        dx = dy = dz = x[1] - x[0]
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        f = Y**3 * X**3 * Z**3
        fxe = 3 * X**2 * Y**3 * Z**3
        fx = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[0])(f)
        err = np.max(np.abs(fxe - fx))
        self.assertAlmostEqual(0, err, 4)
        fye = 3 * X**3 * Y**2 * Z**3
        fy = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[1])(f)
        err = np.max(np.abs(fye - fy))
        self.assertAlmostEqual(0, err, 4)
        fze = 3 * X**3 * Y**3 * Z**2
        fz = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[2])(f)
        err = np.max(np.abs(fze - fz))
        self.assertAlmostEqual(0, err, 4)

    def test_diff_3d_ord2_acc2(self):
        x = np.linspace(-0.5, 0.5, 100)
        y = np.linspace(-0.5, 0.5, 100)
        z = np.linspace(-0.5, 0.5, 100)
        dx = dy = dz = x[1] - x[0]
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        f = Y**3 * X**3 * Z**3
        fxxe = 6 * X * Y**3 * Z**3
        fxx = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[0, 0])(f)
        err = np.max(np.abs(fxxe - fxx))
        self.assertAlmostEqual(0, err, 4)
        fyye = 6 * X**3 * Y * Z**3
        fyy = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[1, 1])(f)
        err = np.max(np.abs(fyye - fyy))
        self.assertAlmostEqual(0, err, 6)
        fzze = 6 * X**3 * Y**3 * Z
        fzz = fd.FinDiff(acc=2, h=[dx, dy, dz], dims=[2, 2])(f)
        err = np.max(np.abs(fzze - fzz))
        self.assertAlmostEqual(0, err, 6)

    def test_diff_3d_df_dxdy_acc4(self):
        x = np.linspace(-0.5, 0.5, 100)
        y = np.linspace(-0.5, 0.5, 100)
        z = np.linspace(-0.5, 0.5, 100)
        dx = dy = dz = x[1] - x[0]
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        f = Y**3 * X**3 * Z**3
        fxye = 9 * X**2 * Y**2 * Z**3
        fxy = fd.FinDiff(acc=4, h=[dx, dy, dz], dims=[0, 1])(f)
        err = np.max(np.abs(fxye - fxy))
        self.assertAlmostEqual(0, err, 4)

    def test_diff_3d_df_dxdy_acc6(self):
        x = np.linspace(-12.5, 12.5, 25)
        y = np.linspace(-12.5, 12.5, 25)
        z = np.linspace(-12.5, 12.5, 25)
        dx = dy = dz = x[1] - x[0]
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        f = Y**3 * X**3 * Z**3
        fxye = 9 * X**2 * Y**2 * Z**3
        fxy = fd.FinDiff(acc=4, h=[dx, dy, dz], dims=[0, 1])(f)
        err = np.max(np.abs(fxye - fxy))
        self.assertAlmostEqual(0, err, 4)

    def test_FinDiff_1d(self):
        h, f, fxe, fxxe = self._prep_1d_func()
        d_dx = fd.FinDiff(h=h, dims=[0], acc=4)
        d2_dx2 = fd.FinDiff(h=h, dims=[0, 0], acc=4)
        fx = d_dx(f)
        self._assertAlmostEqual(fx, fxe)
        fxx = d2_dx2(f)
        self._assertAlmostEqual(fxx, fxxe)

    def test_FinDiff_2d(self):
        xy0 = [-5, -5]
        xy1 = [5, 5]
        nxy = [100, 100]
        x = np.linspace(xy0[0], xy1[0], nxy[0])
        y = np.linspace(xy0[1], xy1[1], nxy[1])
        dxy = [x[1] - x[0], y[1] - y[0]]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = X ** 3 * Y ** 3
        fxe = 3 * X ** 2 * Y ** 3
        fyye = 6 * X ** 3 * Y
        d_dx = fd.FinDiff(h=dxy, dims=[0], acc=4)
        fx = d_dx(f)
        d2_dy2 = fd.FinDiff(h=dxy, dims=[1, 1], acc=4)
        fyy = d2_dy2(f)
        self._assertAlmostEqual(fxe, fx)
        self._assertAlmostEqual(fyye, fyy)

    def test_Laplace_2d(self):
        xy0 = [-5, -5]
        xy1 = [5, 5]
        nxy = [100, 100]
        x = np.linspace(xy0[0], xy1[0], nxy[0])
        y = np.linspace(xy0[1], xy1[1], nxy[1])
        dxy = [x[1] - x[0], y[1] - y[0]]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = X ** 3 * Y ** 3
        lap_f_e = 6 * X * Y**3 + 6 * X**3 * Y
        lap = fd.diff.Laplacian(h=dxy)
        lap_f = lap(f)
        self._assertAlmostEqual(lap_f_e, lap_f)


class TestFinDiffNonUniform(unittest.TestCase, TestUtilities):

    def test_BasicFinDiffNonUni_equi(self):
        x = np.linspace(-3, 3, 100)
        f = x**2

        d_dx = fd.diff.BasicFinDiffNonUniform(coords=[x], dims=[0])
        fx = d_dx(f)
        fxe = 2*x
        self._assertAlmostEqual(fx, fxe)

    def test_BasicFinDiffNonUni(self):
        x = np.r_[np.arange(0, 4, 0.005), np.arange(4, 10, 1)]
        f = np.exp(-x**2)

        d_dx = fd.diff.BasicFinDiffNonUniform(coords=[x], dims=[0], acc=2)
        fx = d_dx(f)
        fxe = - 2 * x * np.exp(-x**2)
        self._assertAlmostEqual(fx, fxe, 4)

    def test_BasicFinDiffNonUni_2d(self):
        x = np.r_[np.arange(0, 4, 0.005), np.arange(4, 10, 1)]
        y = np.r_[np.arange(0, 4, 0.005), np.arange(4, 10, 1)]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = np.exp(-X**2-Y**2)

        d_dx = fd.diff.BasicFinDiffNonUniform(coords=[x, y], dims=[0], acc=2)
        fx = d_dx(f)
        fxe = - 2 * X * np.exp(-X**2-Y**2)
        self._assertAlmostEqual(fx, fxe, 4)

    def test_FinDiff_NonUni_2d(self):
        x = np.r_[np.arange(0, 4, 0.005), np.arange(4, 10, 1)]
        y = np.r_[np.arange(0, 4, 0.005), np.arange(4, 10, 1)]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = np.exp(-X**2-Y**2)

        d_dx = fd.FinDiff(coords=[x, y], dims=[0], acc=2)
        fx = d_dx(f)
        fxe = - 2 * X * np.exp(-X**2-Y**2)
        self._assertAlmostEqual(fx, fxe, 4)

    def test_BasicFinDiffNonUni_3d(self):
        x = np.r_[np.arange(0, 4, 0.05), np.arange(4, 10, 1)]
        y = np.r_[np.arange(0, 4, 0.05), np.arange(4, 10, 1)]
        z = np.r_[np.arange(0, 4.5, 0.05), np.arange(4.5, 10, 1)]
        X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
        f = np.exp(-X**2-Y**2-Z**2)

        d_dy = fd.diff.BasicFinDiffNonUniform(coords=[x, y, z], dims=[1], acc=4)
        fy = d_dy(f)
        fye = - 2 * Y * np.exp(-X**2-Y**2-Z**2)
        self._assertAlmostEqual(fy, fye, 4)

    def test_non_uniform_acc_6(self):
        pass


class TestLinearCombinations(unittest.TestCase, TestUtilities):

    def test_addition_of_operators(self):
        xy0 = [-5, -5]
        xy1 = [5, 5]
        nxy = [100, 100]
        x = np.linspace(xy0[0], xy1[0], nxy[0])
        y = np.linspace(xy0[1], xy1[1], nxy[1])
        dxy = [x[1] - x[0], y[1] - y[0]]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = X ** 3 * Y ** 3
        lap_f_e = 6 * X * Y ** 3 + 6 * X ** 3 * Y
        lap = fd.FinDiff(h=dxy, dims=[0, 0]) + fd.FinDiff(h=dxy, dims=[1,1])
        lap_f = lap(f)
        self._assertAlmostEqual(lap_f_e, lap_f)

    def test_multiplication_with_constants(self):
        xy0 = [-5, -5]
        xy1 = [5, 5]
        nxy = [100, 100]
        x = np.linspace(xy0[0], xy1[0], nxy[0])
        y = np.linspace(xy0[1], xy1[1], nxy[1])
        dxy = [x[1] - x[0], y[1] - y[0]]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = X ** 3 * Y**3
        f_diffed_e = 12 * X * Y**3
        diff_op = 2 * fd.FinDiff(h=dxy, dims=[0,0])
        f_diffed = diff_op(f)
        self._assertAlmostEqual(f_diffed, f_diffed_e)

    def test_multiplication_with_variables(self):
        xy0 = [-5, -5]
        xy1 = [5, 5]
        nxy = [100, 100]
        x = np.linspace(xy0[0], xy1[0], nxy[0])
        y = np.linspace(xy0[1], xy1[1], nxy[1])
        dxy = [x[1] - x[0], y[1] - y[0]]
        X, Y = np.meshgrid(x, y, indexing='ij')
        f = X ** 3 * Y**3
        f_diffed_e = 6 * X**2 * Y**3
        diff_op = fd.Coefficient(X) * fd.FinDiff(h=dxy, dims=[0,0])
        f_diffed = diff_op(f)
        self._assertAlmostEqual(f_diffed, f_diffed_e)

    def test_assert_cannot_add_on_incomp_uni_grids(self):

        def do_test():
            xy0 = [-5, -5]
            xy1 = [5, 5]
            nxy = [100, 100]
            x = np.linspace(xy0[0], xy1[0], nxy[0])
            y = np.linspace(xy0[1], xy1[1], nxy[1])
            dxy = [x[1] - x[0], y[1] - y[0]]
            fd.FinDiff(h=dxy, dims=[0, 0]) + fd.FinDiff(h=[1,1], dims=[1, 1])

        self.assertRaises(ValueError, do_test)

    def test_assert_cannot_add_on_uni_and_nonuni_grids(self):

        def do_test():
            xy0 = [-5, -5]
            xy1 = [5, 5]
            nxy = [100, 100]
            x = np.linspace(xy0[0], xy1[0], nxy[0])
            y = np.linspace(xy0[1], xy1[1], nxy[1])
            dxy = [x[1] - x[0], y[1] - y[0]]
            fd.FinDiff(h=dxy, dims=[0, 0]) + fd.FinDiff(coords=[x,y], dims=[1, 1])

        self.assertRaises(ValueError, do_test)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from findiff.diff import FinDiff, Coefficient
from findiff.util import wrap_in_ndarray


class VectorOperator(object):

    def __init__(self, **kwargs):

        if "dims" in kwargs:
            raise ValueError("dims cannot be specified for Gradient")

        if "h" in kwargs:
            h = wrap_in_ndarray(kwargs["h"])
            ndims = len(h)
        if "coords" in kwargs:
            coords = kwargs["coords"]
            if isinstance(coords, np.ndarray):
                shape = coords.shape
                if len(shape) > 1:
                    ndims = shape[0]
                else:
                    ndims = 1
            else:
                ndims = len(coords)

        self.ndims = ndims


class Gradient(VectorOperator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.components = [FinDiff(dims=[k], **kwargs) for k in range(self.ndims)]

    def __call__(self, f):

        if not isinstance(f, np.ndarray):
            raise TypeError("Function to differentiate must be numpy.ndarray")

        if len(f.shape) != self.ndims:
            raise ValueError("Gradients can only be applied to scalar functions")

        result = []
        for k in range(self.ndims):
            d_dxk =  self.components[k]
            result.append(d_dxk(f))

        return np.array(result)


class Divergence(VectorOperator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.components = [FinDiff(dims=[k], **kwargs) for k in range(self.ndims)]

    def __call__(self, f):

        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError("Function to differentiate must be numpy.ndarray or list of numpy.ndarrays")

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError("Divergence can only be applied to vector functions of the same dimension")

        result = np.zeros(f.shape[1:])

        for k in range(self.ndims):
            result += self.components[k](f[k])

        return result


class Curl(VectorOperator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.ndims != 3:
            raise ValueError("Curl operation is only defined in 3 dimensions. {} were given.".format(self.ndims))

        self.components = [FinDiff(dims=[k], **kwargs) for k in range(self.ndims)]

    def __call__(self, f):

        if not isinstance(f, np.ndarray) and not isinstance(f, list):
            raise TypeError("Function to differentiate must be numpy.ndarray or list of numpy.ndarrays")

        if len(f.shape) != self.ndims + 1 and f.shape[0] != self.ndims:
            raise ValueError("Curl can only be applied to vector functions of the three dimensions")

        result = np.zeros(f.shape)

        result[0] += self.components[1](f[2]) - self.components[2](f[1])
        result[1] += self.components[2](f[0]) - self.components[0](f[2])
        result[2] += self.components[0](f[1]) - self.components[1](f[0])

        return result



from math import log
import unittest
import numpy as np
from findiff import FinDiff


class TestScaling(unittest.TestCase):

    def fit_1d(self, acc):
        nx_list = [30, 100, 300, 1000]
        Lx = 10

        log_err_list = []
        log_dx_list = []

        for nx in nx_list:
            x = np.linspace(0, Lx, nx)
            dx = x[1] - x[0]
            f = np.sin(x)
            d_dx = FinDiff(h=dx, dims=[0], acc=acc)
            fx = d_dx(f)
            fxe = np.cos(x)
            err = np.max(np.abs(fxe - fx))
            log_dx_list.append(log(dx))
            log_err_list.append(log(err))

        fit = np.polyfit(log_dx_list, log_err_list, deg=1)
        return fit[0]

    def fit_2d(self, acc):
        nx_list = [10, 30, 100, 300]
        ny_list = [10, 30, 100, 300]
        Lx, Ly = 3, 3

        log_err_list = []
        log_dx_list = []

        for nx, ny in zip(nx_list, ny_list):
            x = np.linspace(0, Lx, nx)
            y = np.linspace(0, Ly, ny)
            dx, dy = x[1] - x[0], y[1] - y[0]
            X, Y = np.meshgrid(x, y, indexing='ij')
            f = np.sin(X) * np.sin(Y)
            d_dx = FinDiff(h=[dx, dy], dims=[0], acc=acc)
            fx = d_dx(f)
            fxe = np.cos(X) * np.sin(Y)
            err = np.max(np.abs(fxe - fx))
            log_dx_list.append(log(dx))
            log_err_list.append(log(err))

        fit = np.polyfit(log_dx_list, log_err_list, deg=1)
        return fit[0]

    def test_1d_acc2(self):
        self.assertAlmostEqual(2, self.fit_1d(acc=2), 1)

    def test_1d_acc4(self):
        self.assertAlmostEqual(4, self.fit_1d(acc=4), 1)

    def test_1d_acc6(self):
        self.assertAlmostEqual(6, self.fit_1d(acc=6), 1)

    def test_2d_acc2(self):
        self.assertAlmostEqual(2, self.fit_2d(acc=2), 1)

    def test_2d_acc4(self):
        self.assertAlmostEqual(4, self.fit_2d(acc=4), 1)


if __name__ == '__main__':
    unittest.main()


from setuptools import setup

setup(
    name='findiff',
    version='0.4.0',
    description='A Python package for finite difference derivatives in any number of dimensions.',
    long_description="""A Python package for finite difference derivatives in any number of dimensions.
    
    Features: 

        * Differentiate arrays of any number of dimensions along any axis
        * Partial derivatives of any desired order
        * Standard operators from vector calculus like gradient, divergence and curl
        * Can handle uniform and non-uniform grids
        * Can handle arbitrary linear combinations of derivatives with constant and variable coefficients
        * Accuracy order can be specified
        * Calculate raw finite difference coefficients for any order and accuracy for uniform and non-uniform grids

    """,

    license='MIT',
    url='https://github.com/maroba/findiff',

    author='Matthias Baer',  # Optional
    author_email='mrbaer@t-online.de',  # Optional

    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Mathematics',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['finite-differences',  'numerical-derivatives', 'scientific-computing'],  # Optional
    packages=['findiff'],
    install_requires=['numpy'],  # Optional

)



import numpy as np
from findiff import *
from timeit import timeit

nx, ny, nz = 100, 100, 100


def stat_timeit(block, number=10):
    times = []
    for _ in range(number):
        times.append(timeit(block, number=1))
    times = np.array(times)
    return np.mean(times), np.min(times), np.max(times), np.std(times)


def non_uniform():
    x = np.r_[np.linspace(0,3, nx/2, endpoint=False), np.linspace(3, 10, nx/2)]
    y = np.r_[np.linspace(0,3, nx/2, endpoint=False), np.linspace(3, 10, nx/2)]
    z = np.r_[np.linspace(0,3, nx/2, endpoint=False), np.linspace(3, 10, nx/2)]
    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
    f = np.exp(-X ** 2 - Y ** 2 - Z ** 2)

    d_dx = FinDiff(coords=[x, y, z], dims=[0], acc=2)
    d_dx(f)


def uniform():

    x = np.linspace(0, 10, nx)
    y = np.linspace(0, 10, ny)
    z = np.linspace(0, 10, nz)

    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
    f = np.exp(-X ** 2 - Y ** 2 - Z ** 2)

    d_dx = FinDiff(h=[x[1]-x[0], y[1]-y[0], z[1]-z[0]], dims=[0], acc=2)
    d_dx(f)


print(stat_timeit(non_uniform, 20))
print(stat_timeit(uniform, 20))